#include "vstPlugin.h"
#include "vst.h"
#include <tchar.h>
#include <stdio.h>
#include <new>
#include "../resource.h"

struct VSTDLGTEMPLATE: DLGTEMPLATE
{
	WORD ext[3];
	VSTDLGTEMPLATE(void)
	{
		memset(this, 0, sizeof(*this));
	}
};

// O[oϐ
CVstPluginLoader*	g_pLoader			= NULL;
AEffect*			g_pEffect			= NULL;
HWND				g_hEditWnd			= NULL;
float*				g_aSource[2]		= {NULL, NULL};
float*				g_aTarget[2]		= {NULL, NULL};
int					g_nMaxSampleCount	= 0;
int					g_nBlockSize		= 0;
float				g_fSampleRate		= 44100.f;

// vg^Cv
static VstIntPtr VSTCALLBACK VST_HostCallback(AEffect* g_pEffect, VstInt32 opcode, VstInt32 index, VstIntPtr value, void* ptr, float opt);
static INT_PTR CALLBACK VST_EditorProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);

//-------------------------------------------------------------------------------------------------
// [h
//-------------------------------------------------------------------------------------------------
BOOL VST_Load(LPCTSTR lpszFileName, int nSampleRate)
{
	VST_Unload();

	g_pLoader = new CVstPluginLoader;

	if (!g_pLoader->loadLibrary(lpszFileName))
	{
		VST_Unload();
		return FALSE;
	}

	VstPluginEntryProc mainEntry = g_pLoader->getMainEntry();
	if (!mainEntry)
	{
		VST_Unload();
		return FALSE;
	}

	g_pEffect = mainEntry(VST_HostCallback);
	if (!g_pEffect)
	{
		VST_Unload();
		return FALSE;
	}

	if (g_pEffect->numInputs != 2 || g_pEffect->numOutputs != 2)
	{
		VST_Unload();
		return FALSE;
	}

	g_pEffect->dispatcher(g_pEffect, effOpen, 0, 0, 0, 0);
	g_pEffect->dispatcher(g_pEffect, effSetSampleRate, 0, 0, 0, (float)nSampleRate);

	// ubNTCYݒ
	g_nBlockSize = -1;

	return TRUE;
}

//-------------------------------------------------------------------------------------------------
// A[h
//-------------------------------------------------------------------------------------------------
void VST_Unload(void)
{
	VST_Suspend();

	if (g_hEditWnd)
	{
		SendMessage(g_hEditWnd, WM_CLOSE, 0, 0);
		g_hEditWnd = NULL;
	}

	if (g_pEffect)
	{
		g_pEffect->dispatcher(g_pEffect, effClose, 0, 0, 0, 0);
		g_pEffect = NULL;
	}

	if (g_pLoader)
	{
		delete g_pLoader;
		g_pLoader = NULL;
	}

	delete [] g_aSource[0];
	delete [] g_aSource[1];
	delete [] g_aTarget[0];
	delete [] g_aTarget[1];

	g_aSource[0] = NULL;
	g_aSource[1] = NULL;
	g_aTarget[0] = NULL;
	g_aTarget[1] = NULL;
}

//-------------------------------------------------------------------------------------------------
// GfBbgʂJ
//-------------------------------------------------------------------------------------------------
BOOL VST_OpenEdit(HWND hParentWnd)
{
	if (!g_pEffect)
	{
		return FALSE;
	}

	if ((g_pEffect->flags & effFlagsHasEditor) == 0)
	{
		return FALSE;
	}

	if (IsWindow(g_hEditWnd))
	{
		SetForegroundWindow(g_hEditWnd);
		return FALSE;
	}

	VSTDLGTEMPLATE t;	
	t.style = WS_POPUPWINDOW|WS_DLGFRAME|DS_MODALFRAME|DS_CENTER;
	t.cx = 100;
	t.cy = 100;

	g_hEditWnd = CreateDialogIndirectParam(GetModuleHandle(0), &t, hParentWnd, (DLGPROC)VST_EditorProc, (LPARAM)g_pEffect);

	VST_EditShowWindow(SW_SHOW);

	return FALSE;
}

//-------------------------------------------------------------------------------------------------
// GfBbgʂ̕\Ԃ𐧌䂷
//-------------------------------------------------------------------------------------------------
void VST_EditShowWindow(int nCmdShow)
{
	if (IsWindow(g_hEditWnd))
	{
		ShowWindow(g_hEditWnd, nCmdShow);
		UpdateWindow(g_hEditWnd);
	}
}

//-------------------------------------------------------------------------------------------------
// ꎞ~
//-------------------------------------------------------------------------------------------------
void VST_Suspend(void)
{
	if (!g_pEffect)
	{
		return;
	}

	g_pEffect->dispatcher(g_pEffect, effMainsChanged, 0, 0, 0, 0);
}

//-------------------------------------------------------------------------------------------------
// ꎞ~
//-------------------------------------------------------------------------------------------------
void VST_Resume(void)
{
	if (!g_pEffect)
	{
		return;
	}

	g_pEffect->dispatcher(g_pEffect, effMainsChanged, 0, 1, 0, 0);
}

//-------------------------------------------------------------------------------------------------
// ő̓Tvݒ肵A蓖Ă
//-------------------------------------------------------------------------------------------------
void VST_SetMaxSampleCount(int nCount)
{
	VST_Suspend();

	g_nMaxSampleCount = nCount;

	delete [] g_aSource[0];
	delete [] g_aSource[1];
	delete [] g_aTarget[0];
	delete [] g_aTarget[1];

	g_aSource[0] = new float [g_nMaxSampleCount];
	g_aSource[1] = new float [g_nMaxSampleCount];
	g_aTarget[0] = new float [g_nMaxSampleCount];
	g_aTarget[1] = new float [g_nMaxSampleCount];

	// [hĂԂłΔf
	if (g_pEffect)
	{
		g_nBlockSize = g_nMaxSampleCount;
		g_pEffect->dispatcher(g_pEffect, effSetBlockSize, 0, g_nBlockSize, 0, 0);
	}

	VST_Resume();
}

//-------------------------------------------------------------------------------------------------
// 
//-------------------------------------------------------------------------------------------------
static inline short VST_Clip(int v)
{
	if (v > 32767)
		return  32767;
	else if (v < -32768)
		return -32768;
	return (short)v;
}

//-------------------------------------------------------------------------------------------------
// M 16bit int Stereo
//-------------------------------------------------------------------------------------------------
BOOL VST_Process(short* pSource, short* pTarget, DWORD dwSampleCount)
{
	if (!g_pEffect || !dwSampleCount)
	{
		return FALSE;
	}

	if (!g_aSource[0] || !g_aSource[1] || !g_aTarget[0] || !g_aTarget[1])
	{
		return FALSE;
	}

	// ubNTCYݒ
	if (g_nBlockSize != g_nMaxSampleCount)
	{
		g_nBlockSize = g_nMaxSampleCount;
		g_pEffect->dispatcher(g_pEffect, effSetBlockSize, 0, g_nBlockSize, 0, 0);
	}

	DWORD i, j;

	// 16bit int -> 32bit float
	for (i = 0, j = 0; i < dwSampleCount; i++, j += 2)
	{
		g_aSource[0][i] = pSource[j + 0] / 32768.0f;	// L
		g_aSource[1][i] = pSource[j + 1] / 32768.0f; 	// R
	}

	g_pEffect->processReplacing(g_pEffect, g_aSource, g_aTarget, (int)dwSampleCount);

	// 32bit float -> 16bit int
	for (i = 0, j = 0; i < dwSampleCount; i++, j += 2)
	{
		pTarget[j + 0] = VST_Clip((int)(g_aTarget[0][i] * 32768.0f + 0.5f));	// L
		pTarget[j + 1] = VST_Clip((int)(g_aTarget[1][i] * 32768.0f + 0.5f));	// R
	}

	return TRUE;
}

//-------------------------------------------------------------------------------------------------
// GfBbg
//-------------------------------------------------------------------------------------------------
INT_PTR CALLBACK VST_EditorProc (HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg)
	{
	//-----------------------
	case WM_INITDIALOG :
		{
			SetWindowText(hwnd, _T("++"));
			SetTimer(hwnd, 1, 20, 0);

			HICON hIcon = LoadIcon(GetModuleHandle(0), MAKEINTRESOURCE(IDI_ICON1));//
			SendMessage(hwnd, WM_SETICON, ICON_BIG, (LPARAM)hIcon);

			if (!g_pEffect)
			{
				break;
			}

			g_pEffect->dispatcher (g_pEffect, effEditOpen, 0, 0, hwnd, 0);

			ERect* eRect = 0;
			g_pEffect->dispatcher(g_pEffect, effEditGetRect, 0, 0, &eRect, 0);

			if (eRect)
			{
				int width = eRect->right - eRect->left;
				int height = eRect->bottom - eRect->top;
				if (width < 100)
					width = 100;
				if (height < 100)
					height = 100;

				RECT wRect;
				SetRect (&wRect, 0, 0, width, height);
				AdjustWindowRectEx (&wRect, GetWindowLong (hwnd, GWL_STYLE), FALSE, GetWindowLong (hwnd, GWL_EXSTYLE));
				width = wRect.right - wRect.left;
				height = wRect.bottom - wRect.top;

				SetWindowPos(hwnd, HWND_TOP, 0, 0, width, height, SWP_NOMOVE);
			}
		}	
		break;
	
	//-----------------------
	case WM_TIMER :
		if (g_pEffect)
			g_pEffect->dispatcher(g_pEffect, effEditIdle, 0, 0, 0, 0);
		break;

	//-----------------------
	case WM_CLOSE :
		KillTimer (hwnd, 1);

		if (g_pEffect)
			g_pEffect->dispatcher(g_pEffect, effEditClose, 0, 0, 0, 0);

		DestroyWindow(hwnd);
		break;

	case WM_DESTROY:
		g_hEditWnd = NULL;
		break;
	}

	return 0;
}

//-------------------------------------------------------------------------------------------------
// R[obN
//-------------------------------------------------------------------------------------------------
VstIntPtr VSTCALLBACK VST_HostCallback(AEffect* g_pEffect, VstInt32 opcode, VstInt32 index, VstIntPtr value, void* ptr, float opt)
{
	VstIntPtr result = 0;

	switch (opcode)
	{
	case audioMasterVersion:
		result = kVstVersion;
		break;

	case audioMasterGetSampleRate:
		if (g_pEffect)
			g_pEffect->dispatcher(g_pEffect, effSetSampleRate, 0, 0, 0, g_fSampleRate);
		break;

	case audioMasterGetBlockSize:
		if (g_pEffect && 0 < g_nBlockSize)
			g_pEffect->dispatcher(g_pEffect, effSetBlockSize, 0, g_nBlockSize, NULL, 0);
		break;

	case audioMasterGetCurrentProcessLevel:
		result = 2; // currently in audio thread (where process is called
		break;
	}

	return result;
}
